"""
Limited-memory BFGS (L-BFGS)

REF: Algorithm 7.5 in Numerical Optimization by Nocedal and Wright
"""

import logging
from typing import Callable, cast

import torch
from torch import Tensor

from sqnm.line_search import strong_wolfe_line_search
from sqnm.optim.sqn_base import SQNBase

logger = logging.getLogger(__name__)


class LBFGS(SQNBase):
    def __init__(
        self,
        params,
        lr: float = 1e-3,
        line_search_fn: str | None = None,
        history_size: int = 20,
        grad_tol: float = 1e-4,
    ):
        """
        Limited-memory BFGS (L-BFGS)

        Parameters:
            params: iterable of parameters to optimize
            lr: learning rate, ignored if line_search_fn is not None
            line_search_fn: line search function to use, either None for fixed step
                size, or "strong_wolfe" for strong Wolfe line search
            history_size: history size, usually 2 <= m <= 30
            grad_tol: termination tolerance for gradient norm
        """
        if line_search_fn is not None and line_search_fn != "strong_wolfe":
            raise ValueError("L-BFGS only supports strong Wolfe line search")

        defaults = dict(
            lr=lr,
            line_search_fn=line_search_fn,
            history_size=history_size,
            grad_tol=grad_tol,
        )
        super().__init__(params, defaults)

    @torch.no_grad()
    def step(  # type: ignore[override]
        self,
        closure: Callable[[], Tensor],
        fn: Callable[[Tensor], Tensor] | None = None,
    ) -> Tensor:
        """
        Perform a single L-BFGS iteration.

        Parameters:
            closure: A closure that re-evaluates the model and returns the loss.
            fn: A pure function that computes the loss for a given input.
        """
        # Get state and hyperparameter variables
        group = self.param_groups[0]
        lr = group["lr"]
        m = group["history_size"]
        grad_tol = group["grad_tol"]
        line_search_fn = group["line_search_fn"]

        state = self.state[self._params[0]]
        k: int = state["num_iters"]
        s_hist: Tensor = state["s_hist"]
        y_hist: Tensor = state["y_hist"]
        alphas: list[float] = state["alphas"]
        sdotys: list[Tensor] = state["sdotys"]

        if line_search_fn == "strong_wolfe" and fn is None:
            raise ValueError("fn parameter is needed for strong Wolfe line search")

        ################################################################################

        # Make sure the closure is always called with grad enabled
        closure = torch.enable_grad()(closure)

        orig_loss = closure()  # Populate gradients
        loss = float(orig_loss)

        xk = self._get_param_vector()
        gradk = self._get_grad_vector()

        if gradk.norm() < grad_tol:
            return orig_loss

        if k == 0:
            pk = -gradk  # Gradient descent for first iteration
        else:
            pk = -self._two_loop_recursion(gradk)
            if gradk.dot(pk) >= 0:
                logger.warning("p_k is not a descent direction.")

        if line_search_fn == "strong_wolfe":
            fn = cast(Callable[[Tensor], Tensor], fn)
            grad_and_val_fn = torch.func.grad_and_value(fn)
            alpha_k, ls_func_evals = strong_wolfe_line_search(
                grad_and_val_fn, xk, pk, loss, gradk
            )
        else:
            # Use fixed step size
            alpha_k = lr

        # Compute and store next iterates
        sk = alpha_k * pk
        self._add_param_vector(sk)
        closure()  # Recompute gradient after setting new param
        gradk_next = self._get_grad_vector()
        yk = gradk_next - gradk

        # Update state
        alphas.append(alpha_k)
        sdotys.append(sk.dot(yk))
        s_hist[state["num_sy_pairs"] % m] = sk
        y_hist[state["num_sy_pairs"] % m] = yk
        state["num_sy_pairs"] += 1
        state["num_iters"] += 1

        return orig_loss



"""
Multi-batch BFGS (MB-BFGS)

Berahas, A. S., Nocedal, J., & Takáč, M. (2016). A Multi-Batch L-BFGS Method for Machine
    Learning.
"""

import logging
from typing import Any, Callable, cast

import torch
from torch import Tensor

from sqnm.line_search import prob_line_search, strong_wolfe_line_search
from sqnm.optim.sqn_base import SQNBase

logger = logging.getLogger(__name__)


class MBBFGS(SQNBase):
    LINE_SEARCH_FNS = ["strong_wolfe", "prob_wolfe"]
    EXCEPTION_STRATS = ["skip_update"]

    def __init__(
        self,
        params,
        batch_size: int,
        overlap_batch_size: int,
        lr: float = 1e-3,
        line_search_fn: str | None = None,
        history_size: int = 20,
        weight_decay: float = 0.0,
        exception_strat: str | None = None,
        eps: float = 1e-8,
    ):
        """
        Multi-Batch BFGS (MB-BFGS)

        Parameters:
            params: iterable of parameters to optimize
            batch_size: used to keep track of number of function evaluations
            overlap_batch_size: used to keep track of number of function evaluations
            lr: learning rate, ignored if line_search_fn is not None
            line_search_fn: line search function to use, either None for fixed step
                size, or one of MBBFGS.LINE_SEARCH_FNS
            history_size: history size, usually 2 <= m <= 30
            weight_decay: l2 regularisation term
            exception_strat: strategy to employ if curvature condition is violated,
                either None for no strategy or one of MBBFGS.EXCEPTION_STRATS
            eps: if exception_strat == "skip_update", the update is skipped if
                s.dot(y) < eps * ||s||^2
        """
        if line_search_fn is not None and line_search_fn not in self.LINE_SEARCH_FNS:
            raise ValueError(f"MB-BFGS only supports one of: {self.LINE_SEARCH_FNS}")
        if exception_strat is not None and exception_strat not in self.EXCEPTION_STRATS:
            raise ValueError(f"MB-BFGS only supports one of: {self.EXCEPTION_STRATS}")

        defaults = dict(
            batch_size=batch_size,
            overlap_batch_size=overlap_batch_size,
            lr=lr,
            line_search_fn=line_search_fn,
            history_size=history_size,
            weight_decay=weight_decay,
            exception_strat=exception_strat,
            eps=eps,
        )
        super().__init__(params, defaults)

    @torch.no_grad()
    def step(  # type: ignore[override]
        self,
        closure: Callable[[], Tensor],
        overlap_fn: Callable[[Tensor], Tensor],
        fn: Callable[[Tensor], Tensor] | Callable[[Tensor, bool], Any] | None = None,
    ):
        # Get state and hyperparameter variables
        group = self.param_groups[0]
        batch_size = group["batch_size"]
        overlap_batch_size = group["overlap_batch_size"]
        lr = group["lr"]
        line_search_fn = group["line_search_fn"]
        m = group["history_size"]
        weight_decay = group["weight_decay"]
        exception_strat = group["exception_strat"]
        eps = group["eps"]

        state = self.state[self._params[0]]
        k: int = state["num_iters"]
        s_hist: Tensor = state["s_hist"]
        y_hist: Tensor = state["y_hist"]
        alphas: list[float] = state["alphas"]
        sdotys: list[Tensor] = state["sdotys"]

        if line_search_fn is not None and fn is None:
            raise ValueError("fn parameter is needed for line search")

        ################################################################################

        # Make sure the closure is always called with grad enabled
        closure = torch.enable_grad()(closure)

        orig_loss = closure()  # Populate gradients
        loss = float(orig_loss)
        state["func_evals"] += 2 * batch_size  # Forward + backward pass

        xk = self._get_param_vector()
        gradk = self._get_grad_vector()
        if weight_decay != 0:
            gradk.add_(xk, alpha=weight_decay)

        # NOTE: Termination criterion?

        if k == 0:
            pk = -gradk  # Gradient descent for first iteration
        else:
            # NOTE: Can't reliably check if pk is a descent direction here
            pk = -self._two_loop_recursion(gradk)

        if line_search_fn == "strong_wolfe":
            fn = cast(Callable[[Tensor], Tensor], fn)
            grad_and_val_fn = torch.func.grad_and_value(fn)
            alpha_k, ls_func_evals = strong_wolfe_line_search(
                grad_and_val_fn, xk, pk, loss, gradk
            )
            state["func_evals"] += ls_func_evals * batch_size
        elif line_search_fn == "prob_wolfe":
            fn = cast(Callable[[Tensor, bool], Any], fn)
            grad_and_val_fn = torch.func.grad_and_value(lambda x: fn(x, False))
            var_f0, var_df0 = fn(xk, True)
            # Don't need function handle to return variances in line search
            alpha_k, _, _, ls_func_evals = prob_line_search(
                grad_and_val_fn, xk, pk, loss, gradk, var_f0, var_df0
            )
            state["func_evals"] += ls_func_evals * batch_size
        else:
            # Use fixed step size
            alpha_k = lr

        sk = alpha_k * pk
        xk_next = xk + sk
        grad_overlap_fn = torch.func.grad(overlap_fn)
        yk = grad_overlap_fn(xk_next) - grad_overlap_fn(xk)
        state["func_evals"] += 4 * overlap_batch_size  # Forward + backward pass

        if exception_strat == "skip_update" and sk.dot(yk) < eps * sk.dot(sk):
            # Curvature condition violated, skip this update
            return orig_loss

        self._add_param_vector(sk)

        # Update state
        alphas.append(alpha_k)
        sdotys.append(sk.dot(yk))
        s_hist[state["num_sy_pairs"] % m] = sk
        y_hist[state["num_sy_pairs"] % m] = yk
        state["num_sy_pairs"] += 1
        state["num_iters"] += 1

        return orig_loss



from typing import Iterable

import torch
import torch.nn as nn
from torch import Tensor


def unflatten(
    flat_vec: Tensor, param_shapes: dict[str, torch.Size]
) -> dict[str, Tensor]:
    out: dict[str, Tensor] = dict()
    offset = 0
    for name, shape in param_shapes.items():
        numel = int(torch.prod(torch.tensor(shape)).item())
        out[name] = flat_vec[offset : offset + numel].view(shape)
        offset += numel
    return out


def flatten(params: dict) -> Tensor:
    return torch.cat([param.flatten() for param in params.values()])


def grad_vec(params: Iterable[nn.Parameter]) -> Tensor:
    """Concatenates gradients from params into a 1D tensor"""
    grads = []
    for param in params:
        if param.grad is not None:
            grads.append(param.grad.view(-1))
    return torch.cat(grads)



from sqnm.optim.lbfgs import LBFGS
from sqnm.optim.mbbfgs import MBBFGS
from sqnm.optim.olbfgs import OLBFGS
from sqnm.optim.scbfgs import SCBFGS
from sqnm.optim.sgd_with_ls import SGDWithLS
from sqnm.optim.sqnhv import SQNHv

__all__ = ["LBFGS", "MBBFGS", "OLBFGS", "SCBFGS", "SGDWithLS", "SQNHv"]



import logging
from typing import Callable

from torch import Tensor

logger = logging.getLogger(__name__)


def armijo_line_search(
    fn: Callable[[Tensor], Tensor],
    xk: Tensor,
    pk: Tensor,
    f_xk: float,
    grad_f_xk: Tensor,
    a0: float = 1,
    c: float = 1e-4,
) -> tuple[float, int]:
    """
    Finds an optimal step size that the Armijo/sufficient decrease condition using a
    backtracking strategy, alnog with the number of function evaluations used.

    Parameters:
        fn: objective function, assumed to be bounded below along the direction p_k
        xk: current iterate
        pk: direction, assumed to be a descent direction
        f_xk: initial function value at xk
        grad_f_xk: initial gradient value at xk
        a0: initial step size
        c: parameter for Armijo/sufficient decrease condition

    REF: Algorithm 3.1 in Numerical Optimization by Nocedal and Wright
    """
    ls_func_evals = 0

    grad_f_xk_dot_pk = grad_f_xk.dot(pk)
    a_curr = a0

    while fn(xk + a_curr * pk) > f_xk + c * a_curr * grad_f_xk_dot_pk:
        ls_func_evals += 1
        a_curr /= 2
    ls_func_evals += 1
    return a_curr, ls_func_evals



"""
Self-correcting BFGS (SC-BFGS)

Curtis, F. E. (2016). A Self-Correcting Variable-Metric Algorithm for Stochastic
    Optimization.
"""

import logging
from typing import Any, Callable, cast

import numpy as np
import torch
from torch import Tensor

from sqnm.line_search import prob_line_search, strong_wolfe_line_search
from sqnm.optim.sqn_base import SQNBase

logger = logging.getLogger(__name__)


class SCBFGS(SQNBase):
    LINE_SEARCH_FNS = ["strong_wolfe", "prob_wolfe"]

    def __init__(
        self,
        params,
        batch_size: int,
        lr: float = 1e-3,
        line_search_fn: str | None = None,
        history_size: int = 20,
        stable: bool = True,
        eta1: float = 1 / 4,
        eta2: float = 4,
        rho: float = 1 / 8,
        tau: float = 8,
        weight_decay: float = 0.0,
    ):
        """
        Self-Correcting BFGS (SC-BFGS)

        Parameters:
            params: iterable of parameters to optimize
            batch_size: used to keep track of number of function evaluations
            lr: learning rate, ignored if line_search_fn is not None
            line_search_fn: line search function to use, either None for fixed step
                size, or one of SCBFGS.LINE_SEARCH_FNS
            history_size: history size, usually 2 <= m <= 30
            stable: if True, the current update is skipped if self-correcting conditions
                aren't met for an independent proxy batch (skips up to two batches)
            eta1: lower-bound for s.dot(y)/||s||^2 in self-correcting theorem
            eta2: upper-bound for ||y||^2/s.dot(y) in self-correcting theorem
            rho: scaling constant in self-correcting assumption
            tau: shift constant in self-correcting assumption
            weight_decay: l2 regularisation term
        """
        if line_search_fn is not None and line_search_fn not in self.LINE_SEARCH_FNS:
            raise ValueError(f"SC-BFGS only supports one of: {self.LINE_SEARCH_FNS}")
        if eta1 <= 0 or eta1 >= 1:
            raise ValueError("eta1 should be in the range (0, 1)")
        if eta2 < 1:
            raise ValueError("eta2 should be in the range [1, inf)")

        defaults = dict(
            batch_size=batch_size,
            lr=lr,
            line_search_fn=line_search_fn,
            history_size=history_size,
            stable=stable,
            eta1=eta1,
            eta2=eta2,
            rho=rho,
            tau=tau,
            weight_decay=weight_decay,
        )
        super().__init__(params, defaults)

        state = self.state[self._params[0]]
        state["num_proxy_funcs"] = 0

    def _compute_beta(self, sk, yk) -> float:
        group = self.param_groups[0]
        eta1 = group["eta1"]
        eta2 = group["eta2"]

        sksk = sk.dot(sk)
        skyk = sk.dot(yk)
        ykyk = yk.dot(yk)

        beta_lower = 0
        if skyk < eta1 * sksk:
            beta_lower = (eta1 * sksk - skyk) / (sksk - skyk)
        yy = beta_lower * sk + (1 - beta_lower) * yk
        if beta_lower > 0 and yy.dot(yy) > eta2 * sk.dot(yy):
            beta_lower = 1
        beta_upper = 0
        if ykyk > eta2 * skyk:
            # ax^2 + bx + c
            a = (sksk - 2 * skyk + ykyk).item()
            b = (2 * skyk - 2 * ykyk - eta2 * sksk + eta2 * skyk).item()
            c = (ykyk - eta2 * skyk).item()
            if np.all(np.isfinite([a, b, c])):
                beta_upper = np.min(np.roots([a, b, c]))
            else:
                beta_upper = 1

        return max(beta_lower, beta_upper)

    def _check_self_correcting_conditions(
        self, xk: Tensor, pk: Tensor, proxy_fn: Callable[[Tensor], Tensor]
    ) -> bool:
        group = self.param_groups[0]
        rho = group["rho"]
        tau = group["tau"]

        proxy_grad_fn = torch.func.grad(proxy_fn)
        proxy_gradk = proxy_grad_fn(xk)
        norm_sq = proxy_gradk.dot(proxy_gradk)
        return rho * norm_sq <= -proxy_gradk.dot(pk) and pk.dot(pk) <= tau * norm_sq

    @torch.no_grad()
    def step(  # type: ignore[override]
        self,
        closure: Callable[[], Tensor],
        proxy_fn: Callable[[Tensor], Tensor] | None = None,
        fn: Callable[[Tensor], Tensor] | Callable[[Tensor, bool], Any] | None = None,
    ):
        # Get state and hyperparameter variables
        group = self.param_groups[0]
        batch_size = group["batch_size"]
        lr = group["lr"]
        line_search_fn = group["line_search_fn"]
        m = group["history_size"]
        stable = group["stable"]
        weight_decay = group["weight_decay"]

        state = self.state[self._params[0]]
        k: int = state["num_iters"]
        s_hist: Tensor = state["s_hist"]
        y_hist: Tensor = state["y_hist"]
        alphas: list[float] = state["alphas"]
        sdotys: list[Tensor] = state["sdotys"]

        if line_search_fn is not None and fn is None:
            raise ValueError("fn parameter is needed for line search")
        if stable and proxy_fn is None:
            raise ValueError("proxy_fn parameter is needed for stable SC-BFGS")

        ################################################################################

        # Make sure the closure is always called with grad enabled
        closure = torch.enable_grad()(closure)

        orig_loss = closure()  # Populate gradients
        loss = float(orig_loss)
        state["func_evals"] += 2 * batch_size  # Forward + backward pass

        xk = self._get_param_vector()
        gradk = self._get_grad_vector()
        if weight_decay != 0:
            gradk.add_(xk, alpha=weight_decay)

        # NOTE: Termination criterion?

        if k == 0:
            pk = -gradk  # Gradient descent for first iteration
        else:
            # Compute (sk, yk) from previous iteration
            # sk computed already - need yk using this iteration's stochastic gradient
            sk = s_hist[state["num_sy_pairs"] % m]
            vk = gradk - state["gradk_prev"]
            yk_tmp = state["alpha_k_prev"] * vk
            betak = self._compute_beta(sk, yk_tmp)
            yk = betak * sk + (1 - betak) * yk_tmp
            if stable:
                # Keep reference to the y we've replaced, in case we need to discard yk
                y_tmp = y_hist[state["num_sy_pairs"] % m]
            # Set new curvature pair (sk, yk)
            alphas.append(state["alpha_k_prev"])
            sdotys.append(sk.dot(yk))
            y_hist[state["num_sy_pairs"] % m] = yk
            state["num_sy_pairs"] += 1

            # NOTE: Can't reliably check if pk is a descent direction here
            pk = -self._two_loop_recursion(gradk)

            if stable:
                assert proxy_fn is not None
                # Check self-correcting assumptions using proxy batch
                sc_cond = self._check_self_correcting_conditions(xk, pk, proxy_fn)
                state["func_evals"] += 2 * batch_size
                state["num_proxy_funcs"] += 1
                if sc_cond:
                    # Assumptions satisfied, reset counter and use this curvature pair
                    state["num_proxy_funcs"] = 0
                else:
                    # Not satisfied, discard this batch (undo curvature pair)
                    alphas.pop()
                    sdotys.pop()
                    y_hist[state["num_sy_pairs"] % m] = y_tmp
                    state["num_sy_pairs"] -= 1
                    # If we haven't reached the max number of evals, try again with a
                    # different batch, else continue
                    if state["num_proxy_funcs"] < 2:
                        return orig_loss

        if line_search_fn == "strong_wolfe":
            fn = cast(Callable[[Tensor], Tensor], fn)
            grad_and_val_fn = torch.func.grad_and_value(fn)
            alpha_k, ls_func_evals = strong_wolfe_line_search(
                grad_and_val_fn, xk, pk, loss, gradk
            )
            state["func_evals"] += ls_func_evals * batch_size
        elif line_search_fn == "prob_wolfe":
            fn = cast(Callable[[Tensor, bool], Any], fn)
            grad_and_val_fn = torch.func.grad_and_value(lambda x: fn(x, False))
            var_f0, var_df0 = fn(xk, True)
            # Don't need function handle to return variances in line search
            alpha_k, _, _, ls_func_evals = prob_line_search(
                grad_and_val_fn, xk, pk, loss, gradk, var_f0, var_df0
            )
            state["func_evals"] += ls_func_evals * batch_size
        else:
            # Use fixed step size
            alpha_k = lr

        sk = alpha_k * pk
        self._add_param_vector(sk)

        # Can only store sk now, need next batch to compute yk
        s_hist[state["num_sy_pairs"] % m] = sk
        state["alpha_k_prev"] = alpha_k
        state["gradk_prev"] = gradk

        state["num_iters"] += 1
        return orig_loss



import logging
from functools import wraps

import torch
from torch import Tensor

from ..utils.matrix import block_tensor

OFFSET = 10.0

logger = logging.getLogger(__name__)


def cached_computation(method):
    key = method.__name__

    @wraps(method)
    def wrapper(self, x):
        if key not in self.func_cache:
            self.func_cache[key] = {}

        if x in self.func_cache[key]:
            return self.func_cache[key][x]

        result = method(self, x)
        self.func_cache[key][x] = result
        return result

    return wrapper


class ProbLSGaussianProcess:
    def __init__(self, sigma_f, sigma_df):
        self.sigma_f = sigma_f
        self.sigma_df = sigma_df

        self.N = 0
        self.ts = []
        self.ys = []
        self.dys = []

        self.K = None
        self.Kd = None
        self.dKd = None

        self.A = None
        self.G = None

        self.func_cache = {}

    def add(self, t, y, dy, var_f=0.0, var_df=0.0):
        self.N += 1
        self.ts.append(t)
        self.ts_vec = torch.as_tensor(self.ts).t()
        self.ys.append(y)
        self.dys.append(dy)

        self.K = _k(self.ts_vec.unsqueeze(1), self.ts_vec.unsqueeze(0))
        self.Kd = _kd(self.ts_vec.unsqueeze(1), self.ts_vec.unsqueeze(0))
        self.dKd = _dkd(self.ts_vec.unsqueeze(1), self.ts_vec.unsqueeze(0))

        noise_vector = torch.tensor([self.sigma_f**2] * (2 * self.N))
        noise_vector[self.N :] = self.sigma_df**2
        k_matrix = block_tensor(self.K, self.Kd, self.Kd.t(), self.dKd)
        self.G = torch.diag(noise_vector) + k_matrix
        # Store LU decomposition of G so we can compute Gx = b
        self.LU, self.LU_pivots = torch.linalg.lu_factor(self.G)

        resid = torch.as_tensor(self.ys + self.dys).unsqueeze(1)
        self.A = torch.linalg.solve(self.G, resid)

        # Reset cache every time the GP is updated
        for func in self.func_cache:
            self.func_cache[func].clear()

    @cached_computation
    def _solve_G(self, b):
        """Solve Gx = b"""
        return torch.linalg.lu_solve(self.LU, self.LU_pivots, b)

    @cached_computation
    def mu(self, t):
        """Posterior mean of GP at t"""
        if not isinstance(t, Tensor):
            t = torch.as_tensor(t)
        T = self.ts_vec
        return _concat(_k(t, T), _kd(t, T)) @ self.A

    @cached_computation
    def d1mu(self, t):
        """First derivative of posterior mean of GP at t"""
        if not isinstance(t, Tensor):
            t = torch.as_tensor(t)
        T = self.ts_vec
        return _concat(_dk(t, T), _dkd(t, T)) @ self.A

    @cached_computation
    def d2mu(self, t):
        """Second derivative of posterior mean of GP at t"""
        if not isinstance(t, Tensor):
            t = torch.as_tensor(t)
        T = self.ts_vec
        return _concat(_ddk(t, T), _ddkd(t, T)) @ self.A

    @cached_computation
    def d3mu(self, t):
        """Third derivative of posterior mean of GP at t"""
        if not isinstance(t, Tensor):
            t = torch.as_tensor(t)
        T = self.ts_vec
        return _concat(_dddk(t, T), torch.zeros(1, self.N)) @ self.A

    @cached_computation
    def V(self, t):
        """Posterior variance of function values at t"""
        if not isinstance(t, Tensor):
            t = torch.as_tensor(t)
        T = self.ts_vec
        k_tt = _k(t, t)
        k_vector = _concat(_k(t, T), _kd(t, T))
        return k_tt - k_vector @ self._solve_G(k_vector.t())

    @cached_computation
    def Vd(self, t):
        """Posterior variance of function values and derivatives at t"""
        if not isinstance(t, Tensor):
            t = torch.as_tensor(t)
        T = self.ts_vec
        kd_tt = _kd(t, t)
        k_vector_1 = _concat(_k(t, T), _kd(t, T))
        k_vector_2 = _concat(_dk(t, T), _dkd(t, T)).t()
        return kd_tt - k_vector_1 @ self._solve_G(k_vector_2)

    @cached_computation
    def dVd(self, t):
        """Posterior variance of derivatives at t"""
        if not isinstance(t, Tensor):
            t = torch.as_tensor(t)
        T = self.ts_vec
        dkd_tt = _dkd(t, t)
        k_vector = _concat(_dk(t, T), _dkd(t, T))
        return dkd_tt - k_vector @ self._solve_G(k_vector.t())

    @cached_computation
    def V0f(self, t):
        """Posterior covariances of function values at t = 0 and t"""
        if not isinstance(t, Tensor):
            t = torch.as_tensor(t)
        T = self.ts_vec
        zero = torch.zeros(1)
        k_0t = _k(zero, t)
        k_vector_1 = _concat(_k(zero, T), _kd(zero, T))
        k_vector_2 = _concat(_k(t, T), _kd(t, T)).t()
        return k_0t - k_vector_1 @ self._solve_G(k_vector_2)

    @cached_computation
    def Vd0f(self, t):
        """Posterior covariance of gradient and function value at t = 0 and t resp."""
        if not isinstance(t, Tensor):
            t = torch.as_tensor(t)
        T = self.ts_vec
        zero = torch.zeros(1)
        dk_0t = _dk(zero, t)
        k_vector_1 = _concat(_dk(zero, T), _dkd(zero, T))
        k_vector_2 = _concat(_k(t, T), _kd(t, T)).t()
        return dk_0t - k_vector_1 @ self._solve_G(k_vector_2)

    @cached_computation
    def V0df(self, t):
        """Posterior covariance of function value and gradient at t = 0 and t resp."""
        if not isinstance(t, Tensor):
            t = torch.as_tensor(t)
        T = self.ts_vec
        zero = torch.zeros(1)
        kd_0t = _kd(zero, t)
        k_vector_1 = _concat(_k(zero, T), _kd(zero, T))
        k_vector_2 = _concat(_dk(t, T), _dkd(t, T)).t()
        return kd_0t - k_vector_1 @ self._solve_G(k_vector_2)

    @cached_computation
    def Vd0df(self, t):
        """Same as _V0f() but for gradients"""
        if not isinstance(t, Tensor):
            t = torch.as_tensor(t)
        T = self.ts_vec
        zero = torch.zeros(1)
        dkd_0t = _dkd(zero, t)
        k_vector_1 = _concat(_dk(zero, T), _dkd(zero, T))
        k_vector_2 = _concat(_dk(t, T), _dkd(t, T)).t()
        return dkd_0t - k_vector_1 @ self._solve_G(k_vector_2)


def _concat(a: Tensor, b: Tensor):
    if a.dim() == 1:
        a = a.unsqueeze(0)
    if b.dim() == 1:
        b = b.unsqueeze(0)
    return torch.cat((a, b), dim=1)


def _k(a, b):
    c = torch.minimum(a + OFFSET, b + OFFSET)
    return (c**3) / 3 + 0.5 * torch.abs(a - b) * (c**2)


def _kd(a, b):
    aa = a + OFFSET
    bb = b + OFFSET
    return ((a < b) * 0.5 * aa**2) + ((a >= b) * aa * bb - 0.5 * (bb**2))


def _dk(a, b):
    return _kd(b, a)


def _dkd(a, b):
    return torch.minimum(a + OFFSET, b + OFFSET)


def _ddk(a, b):
    return (a <= b) * (b - a)


def _ddkd(a, b):
    return (a <= b).float()


def _dddk(a, b):
    return -(a <= b).float()



import logging
from typing import Callable

import numpy as np
from torch import Tensor

logger = logging.getLogger(__name__)


def strong_wolfe_line_search(
    fn: Callable[[Tensor], tuple[Tensor, Tensor]],
    xk: Tensor,
    pk: Tensor,
    f_xk: float,
    grad_f_xk: Tensor,
    a0: float = 1.0,
    a_max: float = 10,
    c1: float = 1e-4,
    c2: float = 0.9,
    max_iters: int = 10,
    zoom_max_iters: int = 10,
) -> tuple[float, int]:
    """
    Finds and returns an optimal step size that satisfies strong Wolfe conditions,
    along with the number of function evaluations used.

    Parameters:
        fn: function to compute a tuple of (gradient, objective) for a given input,
            objective function is assumed to be bounded below along the direction p_k
        xk: current iterate
        pk: direction, assumed to be a descent direction
        f_xk: initial function value at xk
        grad_f_xk: initial gradient value at xk
        a0: initial step size (1 should always be used as the initial step size for
            Newton and quasi-Newton methods)
        a_max: maximum step size
        c1: parameter for Armijo/sufficient decrease condition
        c2: parameter for curvature condition
        max_iters: max number of line search iterations to compute
        zoom_max_iters: max number of zoom() iterations to compute

    REF: Algorithm 3.5 in Numerical Optimization by Nocedal and Wright
    """

    func_evals = 0

    a_prev = 0.0
    a_curr = a0
    a_star = a_curr
    grad_phi0 = grad_f_xk.dot(pk).item()
    phi_prev, grad_phi_prev = f_xk, grad_phi0

    best_armijo_a = a_curr
    best_armijo_phi = float("inf")

    def phi(a_k: float) -> tuple[float, float]:
        grad, val = fn(xk + a_k * pk)
        return grad.dot(pk).item(), val.item()

    def armijo_ok(phi_a: float, a: float) -> bool:
        return phi_a <= f_xk + c1 * a * grad_phi0

    def curvature_ok(grad_phi_a: float) -> bool:
        return abs(grad_phi_a) <= -c2 * grad_phi0

    def zoom(
        a_lo: float,
        a_hi: float,
        phi_lo: float,
        phi_hi: float,
        grad_phi_lo: float,
        grad_phi_hi: float,
    ) -> float:
        """REF: Algorithm 3.6 in Numerical Optimization by Nocedal and Wright"""

        # Maintain three conditions in each iteration:
        # (a) The interval (a_lo, a_hi) contains step lengths that satisfy strong Wolfe
        # (b) a_lo gives the smallest function value among all step lengths generated so
        #     far that satisfy the sufficient decrease condition
        # (c) a_hi is chosen s.t. grad_phi(a_lo) * (a_hi - a_lo) < 0
        nonlocal func_evals

        z_iters = 0
        while z_iters < zoom_max_iters:
            z_iters += 1
            # Interpolate to find a trial step size a_j in (a_lo, a_hi)
            a_j = _cubic_interp(
                a_lo,
                a_hi,
                phi_lo,
                phi_hi,
                grad_phi_lo,
                grad_phi_hi,
            )
            # a_j should be in [a_lo, a_hi]... fallback to the midpoint if not
            if not _inside(a_j, a_lo, a_hi):
                a_j = (a_lo + a_hi) / 2

            grad_phi_j, phi_j = phi(a_j)
            func_evals += 2
            if not armijo_ok(phi_j, a_j) or phi_j >= phi_lo:
                # Narrow search to (a_lo, a_j) - unless a_lo == a_j
                if abs(a_lo - a_j) < 1e-9:
                    break
                a_hi, phi_hi, grad_phi_hi = a_j, phi_j, grad_phi_j
            else:
                if curvature_ok(grad_phi_j):
                    # a_j satisfies strong Wolfe conditions, stop here
                    break
                # Maintain condition (c)
                if grad_phi_j * (a_hi - a_lo) >= 0:
                    a_hi, phi_hi, grad_phi_hi = a_lo, phi_lo, grad_phi_lo
                # Maintain condition (b)
                a_lo, phi_lo, grad_phi_lo = a_j, phi_j, grad_phi_j

        # NOTE: Returning here without satisfying strong Wolfe conditions
        return a_j

    iters = 1
    while iters <= max_iters:
        grad_phi_curr, phi_curr = phi(a_curr)
        func_evals += 2

        armijo = armijo_ok(phi_curr, a_curr)

        if armijo and phi_curr < best_armijo_phi:
            best_armijo_a, best_armijo_phi = a_curr, phi_curr

        if not armijo or (phi_curr >= phi_prev and iters > 1):
            # (a_prev, a_curr) contains step lengths satisfying strong Wolfe conditions
            a_star = zoom(
                a_prev, a_curr, phi_prev, phi_curr, grad_phi_prev, grad_phi_curr
            )
            break

        if curvature_ok(grad_phi_curr):
            # a_curr satisfies strong Wolfe conditions, stop here
            a_star = a_curr
            break

        if grad_phi_curr >= 0:
            # (a_prev, a_curr) contains step lengths satisfying strong Wolfe conditions
            a_star = zoom(
                a_curr, a_prev, phi_prev, phi_curr, grad_phi_prev, grad_phi_curr
            )
            break

        # Extrapolate to find next trial value (doubling strategy)
        a_prev, a_curr = a_curr, min(a_curr * 2, a_max)
        phi_prev, grad_phi_prev = phi_curr, grad_phi_curr
        iters += 1

    # Fallback to best Armijo point if Wolfe conditions aren't satisfied
    if iters > max_iters:
        return best_armijo_a, func_evals
    return a_star, func_evals


def _cubic_interp(
    x1: float, x2: float, f1: float, f2: float, g1: float, g2: float
) -> float:
    """
    Find the minimizer of the Hermite-cubic polynomial interpolating a function
    of one variable, at the two points x1 and x2, using the function values f(x_1) = f1
    and f(x_2) = f2 and derivatives grad_f(x_1) = g2 and grad_f(x_2) = g2.

    REF: Equation 3.59 in Numerical Optimization, Nocedal and Wright
    """
    if x1 == x2:
        return x1
    with np.errstate(divide="raise", over="raise", invalid="raise"):
        try:
            d1 = g1 + g2 - 3 * (f1 - f2) / (x1 - x2)
            discriminant = d1**2 - g1 * g2
            if discriminant < 0:
                # Default to midpoint
                return 0.5 * (x1 + x2)
            d2 = np.sign(x2 - x1) * np.sqrt(discriminant)
            return x2 - (x2 - x1) * (g2 + d2 - d1) / (g2 - g1 + 2 * d2)
        except ArithmeticError:
            # Default to midpoint
            return 0.5 * (x1 + x2)


def _inside(x: float, a: float, b: float) -> bool:
    """Returns whether x is in (a, b)"""
    if not np.isfinite(x):
        return False

    a, b = min(a, b), max(a, b)
    return a <= x <= b



"""
Online limited-memory BFGS (oL-BFGS)

REF: Schraudolph, N. N., Yu, J., & Günter, S. (2007). A Stochastic Quasi-Newton Method
    for Online Convex Optimization.
"""

import logging
from typing import Any, Callable, cast

import torch
from torch import Tensor

from sqnm.line_search import prob_line_search, strong_wolfe_line_search
from sqnm.optim.sqn_base import SQNBase

logger = logging.getLogger(__name__)


class OLBFGS(SQNBase):
    LINE_SEARCH_FNS = ["strong_wolfe", "prob_wolfe"]

    def __init__(
        self,
        params,
        batch_size: int,
        lr: float = 1e-3,
        line_search_fn: str | None = None,
        history_size: int = 20,
        reg_term: float = 0.0,
        c: float = 1.0,
        weight_decay: float = 0.0,
    ):
        """
        Online limited-memory BFGS (oL-BFGS)

        Parameters:
            params: iterable of parameters to optimize
            batch_size: used to keep track of number of function evaluations
            lr: learning rate, ignored if line_search_fn is not None
            line_search_fn: line search function to use, either None for fixed step
                size, or one of OLBFGS.LINE_SEARCH_FNS
            history_size: history size, usually 2 <= m <= 30
            reg_term: regularisation parameter in gradient difference term
            c: step size scaling factor
            weight_decay: l2 regularisation term
        """
        if line_search_fn is not None and line_search_fn not in self.LINE_SEARCH_FNS:
            raise ValueError(f"o-LBFGS only supports one of: {self.LINE_SEARCH_FNS}")

        defaults = dict(
            batch_size=batch_size,
            lr=lr,
            line_search_fn=line_search_fn,
            history_size=history_size,
            reg_term=reg_term,
            c=c,
            weight_decay=weight_decay,
        )
        super().__init__(params, defaults)

    def _two_loop_recursion(self, grad: Tensor) -> Tensor:  # type: ignore[override]
        """
        Two loop recursion for computing H_k * grad

        This differs from the standard two loop recursion in that H_k^(0) is computed
        from the (s, y) pairs from the (up to) history_size most recent iterates,
        instead of just the most recent (s, y) pair.
        """
        group = self.param_groups[0]
        line_search_fn = group["line_search_fn"]
        m = group["history_size"]
        c = group["c"]

        state = self.state[self._params[0]]
        k = state["num_iters"]
        h = min(k, m)  # Number of curvature pairs to use
        idxs = torch.arange(k - h, k) % m
        s = state["s_hist"][idxs]  # [h, d]
        y = state["y_hist"][idxs]  # [h, d]

        q = grad.clone()
        sy = torch.sum(s * y, dim=1)  # [h], precompute s.dot(y) for each pair
        alphas = torch.zeros(h, device=grad.device)
        for i in reversed(range(h)):
            alphas[i] = s[i].dot(q) / sy[i]
            q.sub_(alphas[i] * y[i])
        if line_search_fn is None:
            alphas[-1] *= c  # Scale alpha_{k-1} by c
        r = (torch.sum(sy / torch.sum(y * y, dim=1)) / h) * q
        for i in range(h):
            beta = y[i].dot(r) / sy[i]
            r.add_((alphas[i] - beta) * s[i])
        return r

    @torch.no_grad()
    def step(  # type: ignore[override]
        self,
        closure: Callable[[], Tensor],
        fn: Callable[[Tensor], Tensor] | Callable[[Tensor, bool], Any] | None = None,
    ) -> Tensor:
        """
        Perform a single oL-BFGS iteration.

        Parameters:
            closure: A closure that re-evaluates the model and returns the loss.
            fn: A pure function that computes the loss for a given input. Required if
                using line search. For linesearch_fn == "prob_wolfe", the function
                should also take a boolean parameter which, if True, also returns the
                gradient, loss variance, and gradient variance.
        """
        # Get state and hyperparameter variables
        group = self.param_groups[0]
        batch_size = group["batch_size"]
        lr = group["lr"]
        line_search_fn = group["line_search_fn"]
        m = group["history_size"]
        reg_term = group["reg_term"]
        c = group["c"]
        weight_decay = group["weight_decay"]

        state = self.state[self._params[0]]
        k: int = state["num_iters"]
        s_hist: Tensor = state["s_hist"]
        y_hist: Tensor = state["y_hist"]
        alphas: list[float] = state["alphas"]
        sdotys: list[Tensor] = state["sdotys"]

        if line_search_fn is not None and fn is None:
            raise ValueError("fn parameter is needed for line search")

        ################################################################################

        # Make sure the closure is always called with grad enabled
        closure = torch.enable_grad()(closure)

        orig_loss = closure()  # Populate gradients
        loss = float(orig_loss)
        state["func_evals"] += 2 * batch_size  # Forward + backward pass

        xk = self._get_param_vector()
        gradk = self._get_grad_vector()
        if weight_decay != 0:
            gradk.add_(xk, alpha=weight_decay)

        # NOTE: Termination criterion?

        if k == 0:
            pk = -gradk  # Gradient descent for first iteration
        else:
            # NOTE: Can't reliably check if pk is a descent direction here
            pk = -self._two_loop_recursion(gradk)

        if line_search_fn == "strong_wolfe":
            fn = cast(Callable[[Tensor], Tensor], fn)
            grad_and_val_fn = torch.func.grad_and_value(fn)
            alpha_k, ls_func_evals = strong_wolfe_line_search(
                grad_and_val_fn, xk, pk, loss, gradk
            )
            state["func_evals"] += ls_func_evals * batch_size
        elif line_search_fn == "prob_wolfe":
            fn = cast(Callable[[Tensor, bool], Any], fn)
            grad_and_val_fn = torch.func.grad_and_value(lambda x: fn(x, False))
            var_f0, var_df0 = fn(xk, True)
            # Don't need function handle to return variances in line search
            alpha_k, _, _, ls_func_evals = prob_line_search(
                grad_and_val_fn, xk, pk, loss, gradk, var_f0, var_df0
            )
            state["func_evals"] += ls_func_evals * batch_size
        else:
            # Use fixed step size
            alpha_k = lr / c

        self._add_param_vector(alpha_k * pk)
        closure()  # Recompute gradient after setting new param
        state["func_evals"] += 2 * batch_size  # Forward + backward pass
        gradk_next = self._get_grad_vector()

        sk = alpha_k * pk
        yk = gradk_next - gradk + reg_term * sk

        # Update state
        alphas.append(alpha_k)
        sdotys.append(sk.dot(yk))
        s_hist[state["num_sy_pairs"] % m] = sk
        y_hist[state["num_sy_pairs"] % m] = yk
        state["num_sy_pairs"] += 1
        state["num_iters"] += 1

        return orig_loss



"""
Probabilistic Line Search

REF: Mahsereci, M., & Hennig, P. (2016). Probabilistic Line Searches for Stochastic Optimization.

Maren Mahsereci and Philipp Hennig copyright statement:
***
Copyright (c) 2015 (post NIPS 2015 release 4.0), Maren Mahsereci, Philipp Hennig
mmahsereci@tue.mpg.de, phennig@tue.mpg.de
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the <organization> nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import logging
from typing import Callable

import numpy as np
import torch
from torch import Tensor

from ..utils.bvn import bvn_prob, std_norm_cdf, std_norm_pdf
from ..utils.gaussian_process import ProbLSGaussianProcess

logger = logging.getLogger(__name__)


def prob_line_search(
    fn: Callable[[Tensor], tuple[Tensor, Tensor]],
    x0: Tensor,  # [d]
    dir: Tensor,  # [d]
    f0: float,
    df0: Tensor,  # [d]
    var_f0: float,  # float
    var_df0: Tensor,  # [d]
    a_running_avg: float | None = None,
    a0: float = 1.0,  # QN methods should try a step size of 1.0 first
    L: int = 10,  # max number of function evaluations
    wolfe_threshold: float = 0.3,
) -> tuple[float, float, float, int]:
    ls_func_evals = 0

    grad_phi_0 = df0.dot(dir)
    # Scaling and noise level of GP
    beta = torch.abs(grad_phi_0).item()  # NOTE: df0 not var_df0 as in pseudocode
    if beta <= 1e-12 or np.isnan(beta):
        # Something very wrong, return initial step size
        return _rescale_output(a0, a0, a_running_avg, ls_func_evals)
    sigma_f = np.sqrt(np.float32(var_f0)) / (a0 * beta)
    sigma_df = (torch.sqrt(var_df0.dot(dir**2)) / beta).item()

    gp = ProbLSGaussianProcess(sigma_f, sigma_df)
    tt_ext = 1.0  # Scaled step size for extrapolation
    tt = 1.0  # Scaled step size of first function evaluation (actual is tt * a0)

    gp.add(0.0, 0.0, (grad_phi_0 / beta).item())

    while gp.N < L - 1:
        # Evaluate objective at tt, check if Wolfe prob. is above threshold
        y, dy = _evaluate_objective(fn, f0, tt, x0, a0, dir, beta)
        ls_func_evals += 2
        gp.add(tt, y, dy)
        if _prob_wolfe(tt, gp) > wolfe_threshold:
            return _rescale_output(a0, tt, a_running_avg, ls_func_evals)

        # Find suitable candidates for next evaluation
        ms = [gp.mu(t) for t in gp.ts]
        dms = [gp.d1mu(t) for t in gp.ts]

        min_m_idx = np.argmin(ms)
        m_min, tt_min, dm_min = ms[min_m_idx], gp.ts[min_m_idx], dms[min_m_idx]

        # Check for nearly deterministic (small variance) step size with near zero grad
        if torch.abs(dm_min) < 1e-5 and gp.Vd(tt_min) < 1e-4:
            # Stop here - though Wolfe conditions may not be guaranteed
            y = gp.ys[min_m_idx]
            dy = gp.dys[min_m_idx]
            return _rescale_output(a0, tt, a_running_avg, ls_func_evals)

        # Sort both gp.ts and ms, according to sort order for gp.ts
        ts_sorted, ms_sorted = zip(*sorted(zip(gp.ts, ms)))
        ts_cand = []
        ms_cand = []
        ss_cand = []

        ts_wolfe = []
        ms_wolfe = []

        # Find local minima in all N - 1 cells
        for n in range(gp.N - 1):
            tt_rep = ts_sorted[n] + 1e-6 * (ts_sorted[n + 1] - ts_sorted[n])
            tt_cub_min = _cubic_minimum(tt_rep, gp)

            # If cubic minimum lies in [T_n, T_{n+1}], it is a candidate
            # Otherwise, check if we are going uphill, and break early if so
            if ts_sorted[n] < tt_cub_min < ts_sorted[n + 1]:
                if (
                    not (torch.isnan(tt_cub_min) or torch.isinf(tt_cub_min))
                    and tt_cub_min > 0
                ):
                    ts_cand.append(tt_cub_min.item())
                    ms_cand.append(gp.mu(tt_cub_min))
                    ss_cand.append(torch.sqrt(gp.V(tt_cub_min)))  # NOTE: Take sqrt here
            elif n == 0 and gp.d1mu(0) > 0:
                tt = 0.01 * (ts_sorted[n] + ts_sorted[n + 1])
                return _rescale_output(a0, tt, a_running_avg, ls_func_evals)

            # Check if an old step size is now acceptable
            if n > 0 and _prob_wolfe(ts_sorted[n], gp) > wolfe_threshold:
                ts_wolfe.append(ts_sorted[n])
                ms_wolfe.append(ms_sorted[n])

        # If any points are acceptable, return the step size with the lowest GP mean,
        # with preference for the current step size tt
        if len(ts_wolfe) > 0:
            if tt in ts_wolfe:
                return _rescale_output(a0, tt, a_running_avg, ls_func_evals)
            tt = ts_wolfe[np.argmin(ms_wolfe)]
            return _rescale_output(a0, tt, a_running_avg, ls_func_evals)

        tt_next = max(gp.ts) + tt_ext
        ts_cand.append(tt_next)
        ms_cand.append(gp.mu(tt_next))
        ss_cand.append(torch.sqrt(gp.V(tt_next)))

        # m_min: Collected (not candidate) step size with lowest GP mean
        ei_cand = _expected_improvement(ms_cand, ss_cand, m_min)
        pw_cand = torch.tensor([_prob_wolfe(t, gp) for t in ts_cand])
        t_best_cand = ts_cand[(ei_cand * pw_cand).argmax()]

        # Extend extrapolation step if necessary
        if t_best_cand == tt_next:
            tt_ext *= 2
        tt = t_best_cand

    # Reached budget without finding acceptable point, check final point for acceptance
    y, dy = _evaluate_objective(fn, f0, tt, x0, a0, dir, beta)
    ls_func_evals += 2
    gp.add(tt, y, dy)
    if _prob_wolfe(tt, gp) > wolfe_threshold:
        return _rescale_output(a0, tt, a_running_avg, ls_func_evals)

    # Otherwise - return point with lowest GP mean
    ms = [gp.mu(t) for t in gp.ts[1:]]
    tt = gp.ts[np.argmin(ms) + 1]
    return _rescale_output(a0, tt, a_running_avg, ls_func_evals)


def _rescale_output(
    a0, tt, a_running_avg, ls_func_evals
) -> tuple[float, float, float, int]:
    a_accepted = tt * a0
    if a_running_avg is None:
        return float(a_accepted), -1, -1, ls_func_evals

    a_ext = 1.3
    theta_reset = 100
    gamma = 0.95

    a_running_avg = gamma * a_running_avg + (1 - gamma) * a_accepted
    a_next = a_accepted * a_ext

    # Reset scaling if new GP scaling is drastically different from previous scalings
    if (a_next < a_running_avg / theta_reset) or (a_next > a_running_avg * theta_reset):
        a_next = a_running_avg

    return float(a_accepted), float(a_next), float(a_running_avg), ls_func_evals


def _evaluate_objective(fn, f0, tt, x0, a0, dir, beta):
    """Evaluates and re-scales function and gradient value"""
    x = x0 + tt * a0 * dir
    grad, val = fn(x)
    y = (val - f0) / (a0 * beta)
    dy = (grad @ dir) / beta
    return y, dy


def _cubic_minimum(t, gp: ProbLSGaussianProcess):
    d1m_t = gp.d1mu(t)
    d2m_t = gp.d2mu(t)
    d3m_t = gp.d3mu(t)
    a = 0.5 * d3m_t
    b = d2m_t - t * d3m_t
    c = d1m_t - d2m_t * t + 0.5 * d3m_t * (t**2)

    if torch.abs(d3m_t) < 1e-9:
        return -(d1m_t - t * d2m_t) / d2m_t

    det = b**2 - 4 * a * c
    if det < 0:
        return torch.tensor(torch.inf)

    lr = (-b - torch.sign(a) * torch.sqrt(det)) / (2 * a)
    rr = (-b + torch.sign(a) * torch.sqrt(det)) / (2 * a)

    dt_l = lr - t
    dt_r = rr - t
    cv_l = d1m_t * dt_l + 0.5 * d2m_t * (dt_l**2) + (d3m_t * (dt_l**3)) / 6
    cv_r = d1m_t * dt_r + 0.5 * d2m_t * (dt_r**2) + (d3m_t * (dt_r**3)) / 6

    if cv_l < cv_r:
        return lr
    return rr


def _prob_wolfe(t, gp: ProbLSGaussianProcess, c1=0.05, c2=0.5):
    # Mean and covariance values at start position t = 0
    mu0 = gp.mu(0)
    dmu0 = gp.d1mu(0)
    V0 = gp.V(0)
    Vd0 = gp.Vd(0)
    dVd0 = gp.dVd(0)

    # Marginal mean and variance for Armijo condition
    mu_a = mu0 - gp.mu(t) + c1 * t * dmu0
    V_aa = (
        V0
        + ((c1 * t) ** 2) * dVd0
        + gp.V(t)
        + 2 * (c1 * t * (Vd0 - gp.Vd0f(t)) - gp.V0f(t))
    )

    # Marginal mean and variance for curvature condition
    mu_b = gp.d1mu(t) - c2 * dmu0
    V_bb = (c2**2) * dVd0 - 2 * c2 * gp.Vd0df(t) + gp.dVd(t)

    # Extremely small variances ==> very certain (deterministic evaluation)
    if V_aa <= 1e-9 and V_bb <= 1e-9:
        return float((mu_a >= 0) * (mu_b >= 0))

    # Zero or negative variances (maybe something went wrong?)
    if V_aa <= 0 or V_bb <= 0:
        return 0.0

    # Covariance between conditions
    V_ab = (
        -c2 * (Vd0 + c1 * t * dVd0)
        + c2 * gp.Vd0f(t)
        + gp.V0df(t)
        + c1 * t * gp.Vd0df(t)
        - gp.Vd(t)
    )

    # Noisy case (everything is alright)
    # Correlation
    rho = (V_ab / torch.sqrt(V_aa * V_bb)).item()

    # Lower and upper integral limits for Armijo condition
    low_a = float(-mu_a / torch.sqrt(V_aa))
    up_a = float("inf")

    # Lower and upper integral limits for curvature condition
    low_b = float(-mu_b / torch.sqrt(V_bb))
    b_bar = 2 * c2 * (torch.abs(dmu0) + 2 * torch.sqrt(dVd0))
    up_b = float((b_bar - mu_b) / torch.sqrt(V_bb))

    return bvn_prob(low_a, up_a, low_b, up_b, rho)


def _expected_improvement(m, s, eta):
    m = torch.as_tensor(m)
    s = torch.as_tensor(s)
    m = m.to(s)
    z = (eta - m) / s
    return (eta - m) * std_norm_cdf(z) + s * std_norm_pdf(z)



"""
Base for stochastic quasi-Newton (SQN) optimizers
"""

import logging
from typing import Any

import torch
from torch import Tensor
from torch.nn.utils import parameters_to_vector
from torch.optim import Optimizer
from torch.optim.optimizer import ParamsT

from sqnm.utils.param import grad_vec

logger = logging.getLogger(__name__)


class SQNBase(Optimizer):
    """Base for stochastic quasi-Newton (SQN) optimizers"""

    def __init__(self, params: ParamsT, defaults: dict[str, Any]):
        if "lr" in defaults and defaults["lr"] <= 0:
            raise ValueError("Learning rate must be positive")
        if not defaults.get("history_size"):
            raise ValueError("L-BFGS type optimizers must have a history size")
        if defaults["history_size"] < 1:
            raise ValueError("History size must be positive")

        super().__init__(params, defaults)

        if len(self.param_groups) != 1:
            raise ValueError(
                "L-BFGS type optimizers don't support per-parameter options"
            )

        self._params: list[torch.nn.Parameter] = self.param_groups[0]["params"]
        with torch.no_grad():
            self._flat_params = parameters_to_vector(self._params).detach()

        # Store LBFGS state in first param
        state = self.state[self._params[0]]
        state["num_iters"] = 0
        state["func_evals"] = 0
        state["alphas"] = []
        state["sdotys"] = []
        # Store the m most recent (s, y) pairs
        # s is the iterate difference, y is the gradient difference
        m = defaults["history_size"]
        param = self._get_param_vector()
        state["s_hist"] = torch.zeros(m, param.numel(), device=param.device)
        state["y_hist"] = torch.zeros(m, param.numel(), device=param.device)
        state["num_sy_pairs"] = 0  # Also the next index to insert into

    def _get_grad_vector(self) -> Tensor:
        """Concatenates gradients from all parameters into a 1D tensor"""
        return grad_vec(self._params)

    def _get_param_vector(self) -> Tensor:
        """Returns parameter vector as a flat (1D) tensor"""
        return self._flat_params

    def _add_param_vector(self, vec: torch.Tensor):
        """Add given flat (1D) tensor to model parameters"""
        with torch.no_grad():
            self._flat_params.add_(vec)
            torch.nn.utils.vector_to_parameters(self._flat_params, self._params)

    def _set_param_vector(self, vec: Tensor):
        """Set model parameters to the given flat (1D) tensor"""
        with torch.no_grad():
            self._flat_params.copy_(vec)
            torch.nn.utils.vector_to_parameters(self._flat_params, self._params)

    def _two_loop_recursion(self, grad: Tensor) -> Tensor:
        """
        Two loop recursion for computing H_k * grad

        This may be overridden if a particular optimiser uses a different algorithm,
        e.g. a different starting matrix H_k^(0).

        REF: Algorithm 7.4 in Numerical Optimization by Nocedal and Wright
        """
        group = self.param_groups[0]
        m = group["history_size"]

        state = self.state[self._params[0]]
        k = state["num_sy_pairs"]
        h = min(k, m)  # Number of curvature pairs to use
        idxs = torch.arange(k - h, k) % m
        s = state["s_hist"][idxs]  # [h, d]
        y = state["y_hist"][idxs]  # [h, d]

        q = grad.clone()
        sy = torch.sum(s * y, dim=1)  # [h], precompute s.dot(y) for each pair
        alphas = torch.zeros(h, device=grad.device)
        for i in reversed(range(h)):
            alphas[i] = s[i].dot(q) / sy[i]
            q.sub_(alphas[i] * y[i])
        r = (sy[-1] / (y[-1].dot(y[-1]))) * q
        for i in range(h):
            beta = y[i].dot(r) / sy[i]
            r.add_((alphas[i] - beta) * s[i])
        return r



import torch
from torch import Tensor


def block_tensor(A: Tensor, B: Tensor, C: Tensor, D: Tensor) -> Tensor:
    """
    Creates a block tensor
    A B
    C D
    from tensors A, B, C, D
    """
    top_half, bot_half = torch.cat((A, B, C, D), dim=1).t().chunk(2)
    return torch.cat((top_half, bot_half), dim=1).t()



"""
Hessian-Vector Stochastic Quasi-Newton (SQN-Hv)

REF: Byrd, R. H., Hansen, S. L., Nocedal, J., & Singer, Y. (2015). A Stochastic
    Quasi-Newton Method for Large-Scale Optimization.
"""

import logging
from typing import Any, Callable, cast

import torch
from torch import Tensor
from torch.autograd.functional import hvp

from sqnm.line_search import prob_line_search, strong_wolfe_line_search
from sqnm.optim.sqn_base import SQNBase

logger = logging.getLogger(__name__)


class SQNHv(SQNBase):
    LINE_SEARCH_FNS = ["strong_wolfe", "prob_wolfe"]
    EXCEPTION_STRATS = ["skip_update"]

    def __init__(
        self,
        params,
        batch_size: int,
        curvature_batch_size: int,
        lr: float = 1e-3,
        line_search_fn: str | None = None,
        history_size: int = 20,
        skip: int = 10,
        weight_decay: float = 0.0,
    ):
        """
        Hessian-Vector Stochastic Quasi-Newton (SQN-Hv)

        Parameters:
            params: iterable of parameters to optimize
            batch_size: used to keep track of number of function evaluations
            curvature_batch_size: used to keep track of number of function evaluations
            lr: learning rate, ignored if line_search_fn is not None
            line_search_fn: line search function to use, either None for fixed step
                size, or one of SQNHv.LINE_SEARCH_FNS
            history_size: history size, usually 2 <= m <= 30
            skip: number of iterations between curvature estimates
            weight_decay: l2 regularisation term
        """
        if line_search_fn is not None and line_search_fn not in self.LINE_SEARCH_FNS:
            raise ValueError(f"SQN-Hv only supports one of: {self.LINE_SEARCH_FNS}")

        defaults = dict(
            lr=lr,
            history_size=history_size,
            line_search_fn=line_search_fn,
            skip=skip,
            weight_decay=weight_decay,
            batch_size=batch_size,
            curvature_batch_size=curvature_batch_size,
        )
        super().__init__(params, defaults)

        state = self.state[self._params[0]]
        state["num_iters"] = 1  # Algorithm in paper starts from k = 1
        state["num_sy_pairs"] = -1
        state["xt"] = torch.zeros_like(self._get_param_vector())
        # Used for probabilistic LS
        if line_search_fn == "prob_wolfe":
            state["alpha_start"] = 1.0
            state["alpha_running_avg"] = state["alpha_start"]

    @torch.no_grad()
    def step(  # type: ignore[override]
        self,
        closure: Callable[[], Tensor],
        fn: Callable[[Tensor], Tensor] | Callable[[Tensor, bool], Any] | None = None,
        curvature_fn: Callable[[Tensor], Tensor] | None = None,
    ) -> Tensor:
        """
        Perform a single SQN-Hv iteration.

        Parameters:
            closure: A closure that re-evaluates the model and returns the loss.
            fn: A pure function that computes the loss for a given input. Required if
                using line search. For linesearch_fn == "prob_wolfe", the function
                should also take a boolean parameter which, if True, also returns the
                gradient, loss variance, and gradient variance.
            curvature_fn: A pure function that computes the loss for a given input. The
                function should be provided every `skip` iterations.
        """
        # Get state and hyperparameter variables
        group = self.param_groups[0]
        lr = group["lr"]
        line_search_fn = group["line_search_fn"]
        m = group["history_size"]
        skip = group["skip"]  # L
        weight_decay = group["weight_decay"]
        batch_size = group["batch_size"]
        curvature_batch_size = group["curvature_batch_size"]

        state = self.state[self._params[0]]
        k: int = state["num_iters"]
        s_hist: Tensor = state["s_hist"]
        y_hist: Tensor = state["y_hist"]
        alphas: list[float] = state["alphas"]
        sdotys: list[Tensor] = state["sdotys"]
        # Note index t for curvature pairs, which are decoupled from gradient estimates
        xt: Tensor = state["xt"]

        if line_search_fn is not None and fn is None:
            raise ValueError("fn parameter is needed for line search")

        if k % skip != 0 and curvature_fn is not None:
            logger.warning(f"Got curvature_fn but didn't expect it on iteration {k}")
        if k % skip == 0 and curvature_fn is None:
            raise TypeError(f"Expected curvature_fn but didn't get it on iteration {k}")

        ################################################################################

        # Make sure the closure is always called with grad enabled
        closure = torch.enable_grad()(closure)

        orig_loss = closure()  # Populate gradients
        loss = float(orig_loss)
        state["func_evals"] += 2 * batch_size  # Forward + backward pass

        xk = self._get_param_vector()
        gradk = self._get_grad_vector()
        if weight_decay != 0:
            gradk.add_(xk, alpha=weight_decay)

        # NOTE: Termination criterion?

        if k <= 2 * skip:
            # Stochastic gradient descent for first 2L iterations
            pk = -gradk
        else:
            # NOTE: Can't reliably check if pk is a descent direction here
            pk = -self._two_loop_recursion(gradk)

        if line_search_fn == "strong_wolfe":
            fn = cast(Callable[[Tensor], Tensor], fn)
            grad_and_val_fn = torch.func.grad_and_value(fn)
            alpha_k, ls_func_evals = strong_wolfe_line_search(
                grad_and_val_fn, xk, pk, loss, gradk
            )
            state["func_evals"] += ls_func_evals * batch_size
        elif line_search_fn == "prob_wolfe":
            fn = cast(Callable[[Tensor, bool], Any], fn)
            grad_and_val_fn = torch.func.grad_and_value(lambda x: fn(x, False))
            var_f0, var_df0 = fn(xk, True)
            # Don't need function handle to return vars in line search
            if k <= 2 * skip:
                # Propagate step sizes in probabilistic ls - we don't have curvature
                # information yet
                alpha_k, alpha_start, alpha_running_avg, ls_func_evals = (
                    prob_line_search(
                        grad_and_val_fn,
                        xk,
                        pk,
                        loss,
                        gradk,
                        var_f0,
                        var_df0,
                        a_running_avg=state["alpha_running_avg"],
                        a0=state["alpha_start"],
                    )
                )
                state["alpha_start"] = alpha_start
                state["alpha_running_avg"] = alpha_running_avg
            else:
                alpha_k, _, _, ls_func_evals = prob_line_search(
                    grad_and_val_fn, xk, pk, loss, gradk, var_f0, var_df0
                )
            state["func_evals"] += ls_func_evals * batch_size
        else:
            # Use fixed step size
            alpha_k = lr

        sk = alpha_k * pk
        self._add_param_vector(sk)

        # Accumulate average over L iterations
        xt.add_(xk + sk)

        if k % skip == 0:
            # Compute curvature pairs every L iterations
            t = state["num_sy_pairs"]
            state["num_sy_pairs"] += 1
            xt.div_(skip)
            if t >= 0:
                st = sk
                # Compute subsampled Hessian vector product on a different sample given
                # by curvature_fn
                _, yt = hvp(curvature_fn, xt, v=st, create_graph=False, strict=True)
                state["func_evals"] += 4 * curvature_batch_size
                sdotys.append(st.dot(yt))
                s_hist[t % m] = st
                y_hist[t % m] = yt
            xt.zero_()

        # Update state
        alphas.append(alpha_k)
        state["num_iters"] += 1

        return orig_loss



from sqnm.line_search.armijo_line_search import armijo_line_search
from sqnm.line_search.prob_line_search import prob_line_search
from sqnm.line_search.strong_wolfe_line_search import strong_wolfe_line_search

__all__ = ["armijo_line_search", "prob_line_search", "strong_wolfe_line_search"]



from typing import Callable

import torch
from torch import Tensor
from torch.optim import SGD

from sqnm.line_search import armijo_line_search
from sqnm.utils.param import grad_vec


class SGDWithLS(SGD):
    def __init__(self, batch_size: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        state = self.state[self.param_groups[0]["params"][0]]
        group = self.param_groups[0]
        group["batch_size"] = batch_size
        state["func_evals"] = 0

    def _get_grad_vector(self) -> Tensor:
        """Concatenates gradients from all parameters into a 1D tensor"""
        params = self.param_groups[0]["params"]  # Assume 1 param group
        return grad_vec(params)

    def _get_param_vector(self) -> Tensor:
        """Concatenates all parameters into a 1D tensor"""
        params = self.param_groups[0]["params"]  # Assume 1 param group
        return torch.cat([p.data.view(-1) for p in params])

    @torch.no_grad()
    def step(  # type: ignore[override]
        self,
        closure: Callable[[], float],
        fn: Callable[[Tensor], Tensor],
    ) -> float:
        """
        Perform a single SGD iteration.

        Parameters:
            closure: A closure that re-evaluates the model and returns the loss.
            fn: A pure function that computes the loss for a given input.
        """
        state = self.state[self.param_groups[0]["params"][0]]
        group = self.param_groups[0]
        batch_size = group["batch_size"]

        # Make sure the closure is always called with grad enabled
        closure = torch.enable_grad()(closure)

        orig_loss = closure()  # Populate gradients
        loss = float(orig_loss)
        state["func_evals"] += 2 * batch_size  # Forward + backward pass

        xk = self._get_param_vector()
        gradk = self._get_grad_vector()
        pk = -gradk

        # Choose step size to satisfy Armijo conditions
        alpha_k, ls_func_evals = armijo_line_search(fn, xk, pk, loss, gradk)
        state["func_evals"] += ls_func_evals * batch_size

        # Assume 1 param group
        self.param_groups[0]["lr"] = alpha_k

        # Don't pass closure() to step - we've already populated gradients
        super().step()
        return orig_loss



"""
Translated from a MATLAB implementation by Alan Genz which is based on the method
described by
    Drezner, Z and G.O. Wesolowsky, (1989),
    On the computation of the bivariate normal inegral,
    Journal of Statist. Comput. Simul. 35, pp. 101-107,
with major modifications for double precision, for |r| close to 1.

Alan Genz copyright statement:
***
Copyright (C) 2013, Alan Genz,  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided the following conditions are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the
     distribution.
  3. The contributor name(s) may not be used to endorse or promote
     products derived from this software without specific prior
     written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import numpy as np
from scipy.special import erf  # type: ignore[import-untyped]


def bvn_prob(xl, xu, yl, yu, rho):
    """
    Compute the bivariate normal probability that xl < x < xu and yl < y < yu,
    with mean 0 and covariance matrix [[1, rho], [rho, 1]].
    """
    p = (
        bvn_prob_unbounded(xl, yl, rho)
        - bvn_prob_unbounded(xu, yl, rho)
        - bvn_prob_unbounded(xl, yu, rho)
        + bvn_prob_unbounded(xu, yu, rho)
    )
    return max(0.0, min(1.0, p))


def bvn_prob_unbounded(xl, yl, rho):
    """
    Compute the bivariate normal probability that x > xl and y > yl,
    with mean 0 and covariance matrix [[1, rho], [rho, 1]]
    """
    rho = max(-1.0, min(1.0, rho))
    if np.isposinf(xl) or np.isposinf(yl):
        return 0.0
    if np.isneginf(xl):
        return 1.0 if np.isneginf(yl) else std_norm_cdf(-yl)
    if np.isneginf(yl):
        return std_norm_cdf(-xl)
    if rho == 0.0:
        return std_norm_cdf(-xl) * std_norm_cdf(-yl)

    tp = 2.0 * np.pi
    h = xl
    k = yl
    hk = h * k
    bvn = 0.0

    if np.abs(rho) < 0.3:
        # Gauss Legendre points and weights, n = 6
        w = np.array([0.1713244923791705, 0.3607615730481384, 0.4679139345726904])
        x = np.array([0.9324695142031522, 0.6612093864662647, 0.2386191860831970])
    elif np.abs(rho) < 0.75:
        # Gauss Legendre points and weights, n = 12
        w = np.array(
            [
                0.04717533638651177,
                0.1069393259953183,
                0.1600783285433464,
                0.2031674267230659,
                0.2334925365383547,
                0.2491470458134029,
            ]
        )
        x = np.array(
            [
                0.9815606342467191,
                0.9041172563704750,
                0.7699026741943050,
                0.5873179542866171,
                0.3678314989981802,
                0.1252334085114692,
            ]
        )
    else:
        # Gauss Legendre points and weights, n = 20
        w = np.array(
            [
                0.01761400713915212,
                0.04060142980038694,
                0.06267204833410906,
                0.08327674157670475,
                0.1019301198172404,
                0.1181945319615184,
                0.1316886384491766,
                0.1420961093183821,
                0.1491729864726037,
                0.1527533871307259,
            ]
        )
        x = np.array(
            [
                0.9931285991850949,
                0.9639719272779138,
                0.9122344282513259,
                0.8391169718222188,
                0.7463319064601508,
                0.6360536807265150,
                0.5108670019508271,
                0.3737060887154196,
                0.2277858511416451,
                0.07652652113349733,
            ]
        )

    w = np.tile(w, 2)
    x = np.concatenate([1.0 - x, 1.0 + x])

    if np.abs(rho) < 0.925:
        hs = 0.5 * (h * h + k * k)
        asr = 0.5 * np.arcsin(rho)
        sn = np.sin(asr * x)
        bvn = np.dot(w, np.exp((sn * hk - hs) / (1.0 - sn**2)))
        bvn = bvn * asr / tp + std_norm_cdf(-h) * std_norm_cdf(-k)
    else:
        if rho < 0.0:
            k = -k
            hk = -hk
        if np.abs(rho) < 1.0:
            ass = 1.0 - rho**2
            a = np.sqrt(ass)
            bs = (h - k) ** 2
            asr = -0.5 * (bs / ass + hk)
            c = (4.0 - hk) / 8.0
            d = (12.0 - hk) / 80.0
            if asr > -100.0:
                bvn = (
                    a
                    * np.exp(asr)
                    * (1.0 - c * (bs - ass) * (1.0 - d * bs) / 3.0 + c * d * ass**2)
                )
            if hk > -100.0:
                b = np.sqrt(bs)
                sp = np.sqrt(tp) * std_norm_cdf(-b / a)
                bvn -= (
                    np.exp(-0.5 * hk) * sp * b * (1.0 - c * bs * (1.0 - d * bs) / 3.0)
                )
            a *= 0.5
            xs = (a * x) ** 2
            asr = -0.5 * (bs / xs + hk)
            idxs = (asr > -100.0).nonzero()
            xs = xs[idxs]
            sp = 1.0 + c * xs * (1.0 + 5.0 * d * xs)
            rs = np.sqrt(1.0 - xs)
            ep = np.exp(-0.5 * hk * xs / (1.0 + rs) ** 2) / rs
            bvn = (a * np.dot(np.exp(asr[idxs]) * (sp - ep), w[idxs]) - bvn) / tp
        if rho > 0:
            bvn += std_norm_cdf(-max(h, k))
        elif h >= k:
            bvn = -bvn
        else:
            if h < 0.0:
                L = std_norm_cdf(k) - std_norm_cdf(h)
            else:
                L = std_norm_cdf(-h) - std_norm_cdf(-k)
            bvn = L - bvn
    return max(0.0, min(1.0, bvn))


def std_norm_cdf(z):
    return 0.5 * (1.0 + erf(z / np.sqrt(2.0)))


def std_norm_pdf(z):
    return np.exp(-0.5 * (z**2)) / np.sqrt(2.0 * np.pi)
